"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bignum_1 = __importDefault(require("@liskhq/bignum"));
const lisk_sdk_1 = require("lisk-sdk");
const legacy_base_transaction_1 = require("./legacy_base_transaction");
const { convertToAssetError, TransactionError, utils: { verifyAmountBalance, verifyBalance, }, constants: { BYTESIZES, MAX_TRANSACTION_AMOUNT, TRANSFER_FEE, }, } = lisk_sdk_1.transactions;
const { isValidTransferAmount, isPositiveNumberString, validator, } = lisk_sdk_1.validator;
const { bigNumberToBuffer, hexToBuffer, intToBuffer, stringToBuffer, } = lisk_sdk_1.cryptography;
exports.transferAssetFormatSchema = {
    type: 'object',
    required: ['recipientId', 'amount'],
    properties: {
        recipientId: {
            type: 'string',
            format: 'address',
        },
        amount: {
            type: 'string',
            format: 'amount',
        },
        data: {
            type: 'string',
            format: 'transferData',
            maxLength: 64,
        },
    },
};
class TransferTransaction extends legacy_base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        if (tx.asset) {
            const rawAsset = tx.asset;
            this.asset = {
                data: rawAsset.data,
                recipientId: rawAsset.recipientId,
                amount: new bignum_1.default(isPositiveNumberString(rawAsset.amount) ? rawAsset.amount : '0'),
            };
        }
        else {
            this.asset = {};
        }
    }
    getBasicBytes() {
        const transactionType = intToBuffer(this.type, BYTESIZES.TYPE);
        const transactionTimestamp = intToBuffer(this.timestamp, BYTESIZES.TIMESTAMP, 'little');
        const transactionSenderPublicKey = hexToBuffer(this.senderPublicKey);
        const transactionRecipientID = intToBuffer(this.asset.recipientId.slice(0, -1), BYTESIZES.RECIPIENT_ID).slice(0, BYTESIZES.RECIPIENT_ID);
        const transactionAmount = bigNumberToBuffer(this.asset.amount.toString(), BYTESIZES.AMOUNT, 'little');
        const dataBuffer = this.asset.data
            ? stringToBuffer(this.asset.data)
            : Buffer.alloc(0);
        return Buffer.concat([
            transactionType,
            transactionTimestamp,
            transactionSenderPublicKey,
            transactionRecipientID,
            transactionAmount,
            dataBuffer,
        ]);
    }
    assetToJSON() {
        return {
            data: this.asset.data,
            amount: this.asset.amount.toString(),
            recipientId: this.asset.recipientId,
        };
    }
    async prepare(store) {
        await store.account.cache([
            {
                address: this.senderId,
            },
            {
                address: this.asset.recipientId,
            },
        ]);
    }
    validateAsset() {
        const asset = this.assetToJSON();
        const schemaErrors = validator.validate(exports.transferAssetFormatSchema, asset);
        const errors = convertToAssetError(this.id, schemaErrors);
        if (!isValidTransferAmount(this.asset.amount.toString())) {
            errors.push(new TransactionError('Amount must be a valid number in string format.', this.id, '.amount', this.asset.amount.toString()));
        }
        if (!this.asset.recipientId) {
            errors.push(new TransactionError('`recipientId` must be provided.', this.id, '.asset.recipientId'));
        }
        return errors;
    }
    applyAsset(store) {
        const errors = [];
        const sender = store.account.get(this.senderId);
        const balanceError = verifyAmountBalance(this.id, sender, this.asset.amount, this.fee);
        if (balanceError) {
            errors.push(balanceError);
        }
        const updatedSenderBalance = new bignum_1.default(sender.balance).sub(this.asset.amount);
        const updatedSender = {
            ...sender,
            balance: updatedSenderBalance.toString(),
        };
        store.account.set(updatedSender.address, updatedSender);
        const recipient = store.account.getOrDefault(this.asset.recipientId);
        const updatedRecipientBalance = new bignum_1.default(recipient.balance).add(this.asset.amount);
        if (updatedRecipientBalance.gt(MAX_TRANSACTION_AMOUNT)) {
            errors.push(new TransactionError('Invalid amount', this.id, '.amount', this.asset.amount.toString()));
        }
        const updatedRecipient = {
            ...recipient,
            balance: updatedRecipientBalance.toString(),
        };
        store.account.set(updatedRecipient.address, updatedRecipient);
        return errors;
    }
    undoAsset(store) {
        const errors = [];
        const sender = store.account.get(this.senderId);
        const updatedSenderBalance = new bignum_1.default(sender.balance).add(this.asset.amount);
        if (updatedSenderBalance.gt(MAX_TRANSACTION_AMOUNT)) {
            errors.push(new TransactionError('Invalid amount', this.id, '.amount', this.asset.amount.toString()));
        }
        const updatedSender = {
            ...sender,
            balance: updatedSenderBalance.toString(),
        };
        store.account.set(updatedSender.address, updatedSender);
        const recipient = store.account.getOrDefault(this.asset.recipientId);
        const balanceError = verifyBalance(this.id, recipient, this.asset.amount);
        if (balanceError) {
            errors.push(balanceError);
        }
        const updatedRecipientBalance = new bignum_1.default(recipient.balance).sub(this.asset.amount);
        const updatedRecipient = {
            ...recipient,
            balance: updatedRecipientBalance.toString(),
        };
        store.account.set(updatedRecipient.address, updatedRecipient);
        return errors;
    }
}
exports.TransferTransaction = TransferTransaction;
TransferTransaction.TYPE = 0;
TransferTransaction.FEE = TRANSFER_FEE.toString();
//# sourceMappingURL=0_transfer_transaction.js.map