"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const lisk_sdk_1 = require("lisk-sdk");
const legacy_base_transaction_1 = require("./legacy_base_transaction");
const { convertToAssetError, TransactionError, utils: { getId, }, constants: { SIGNATURE_FEE, }, } = lisk_sdk_1.transactions;
const { hash, hexToBuffer, signData, } = lisk_sdk_1.cryptography;
const { validator, } = lisk_sdk_1.validator;
exports.secondSignatureAssetFormatSchema = {
    type: 'object',
    required: ['publicKey'],
    properties: {
        publicKey: {
            type: 'string',
            format: 'publicKey',
        },
    },
};
class SecondSignatureTransaction extends legacy_base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        this.asset = (tx.asset || { signature: {} });
    }
    assetToBytes() {
        const { publicKey } = this.asset;
        return hexToBuffer(publicKey);
    }
    async prepare(store) {
        await store.account.cache([
            {
                address: this.senderId,
            },
        ]);
    }
    verifyAgainstTransactions(transactions) {
        return transactions
            .filter(tx => tx.type === this.type && tx.senderPublicKey === this.senderPublicKey)
            .map(tx => new TransactionError('Register second signature only allowed once per account.', tx.id, '.asset.signature'));
    }
    validateAsset() {
        const schemaerrors = validator.validate(exports.secondSignatureAssetFormatSchema, this.asset);
        const errors = convertToAssetError(this.id, schemaerrors);
        return errors;
    }
    applyAsset(store) {
        const errors = [];
        const sender = store.account.get(this.senderId);
        if (sender.secondPublicKey) {
            errors.push(new TransactionError('Register second signature only allowed once per account.', this.id, '.secondPublicKey'));
        }
        const updatedSender = {
            ...sender,
            secondPublicKey: this.asset.publicKey,
            secondSignature: 1,
        };
        store.account.set(updatedSender.address, updatedSender);
        return errors;
    }
    undoAsset(store) {
        const sender = store.account.get(this.senderId);
        const resetSender = {
            ...sender,
            secondPublicKey: null,
            secondSignature: 0,
        };
        store.account.set(resetSender.address, resetSender);
        return [];
    }
    sign(passphrase) {
        this._signature = undefined;
        this._signSignature = undefined;
        this._signature = signData(hash(this.getBytes()), passphrase);
        this._id = getId(this.getBytes());
    }
}
exports.SecondSignatureTransaction = SecondSignatureTransaction;
SecondSignatureTransaction.TYPE = 1;
SecondSignatureTransaction.FEE = SIGNATURE_FEE.toString();
//# sourceMappingURL=1_second_signature_transaction.js.map