"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bignum_1 = __importDefault(require("@liskhq/bignum"));
const lisk_sdk_1 = require("lisk-sdk");
const legacy_base_transaction_1 = require("./legacy_base_transaction");
const { convertToAssetError, TransactionError, utils: { verifyAmountBalance, }, constants: { BYTESIZES, MAX_TRANSACTION_AMOUNT, VOTE_FEE, }, } = lisk_sdk_1.transactions;
const { bigNumberToBuffer, hexToBuffer, intToBuffer, stringToBuffer, } = lisk_sdk_1.cryptography;
const { isPositiveNumberString, validator, } = lisk_sdk_1.validator;
const PREFIX_UPVOTE = '+';
const PREFIX_UNVOTE = '-';
const MAX_VOTE_PER_ACCOUNT = 101;
const MIN_VOTE_PER_TX = 1;
const MAX_VOTE_PER_TX = 33;
exports.voteAssetFormatSchema = {
    type: 'object',
    required: ['votes'],
    properties: {
        recipientId: {
            type: 'string',
            format: 'address',
        },
        amount: {
            type: 'string',
            format: 'amount',
        },
        votes: {
            type: 'array',
            minItems: MIN_VOTE_PER_TX,
            maxItems: MAX_VOTE_PER_TX,
            items: {
                type: 'string',
                format: 'signedPublicKey',
            },
            uniqueSignedPublicKeys: true,
        },
    },
};
class VoteTransaction extends legacy_base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        if (tx.asset) {
            const rawAsset = tx.asset;
            this.asset = {
                votes: rawAsset.votes,
                recipientId: rawAsset.recipientId || this.senderId,
                amount: new bignum_1.default(isPositiveNumberString(rawAsset.amount) ? rawAsset.amount : '0'),
            };
        }
        else {
            this.asset = {};
        }
        this.containsUniqueData = true;
    }
    assetToJSON() {
        return {
            votes: this.asset.votes,
            amount: this.asset.amount.toString(),
            recipientId: this.asset.recipientId,
        };
    }
    sign(passphrase, secondPassphrase) {
        super.sign(passphrase, secondPassphrase);
        this.asset.recipientId = this.senderId;
    }
    getBasicBytes() {
        const transactionType = intToBuffer(this.type, BYTESIZES.TYPE);
        const transactionTimestamp = intToBuffer(this.timestamp, BYTESIZES.TIMESTAMP, 'little');
        const transactionSenderPublicKey = hexToBuffer(this.senderPublicKey);
        const transactionRecipientID = intToBuffer(this.asset.recipientId.slice(0, -1), BYTESIZES.RECIPIENT_ID).slice(0, BYTESIZES.RECIPIENT_ID);
        const transactionAmount = bigNumberToBuffer(this.asset.amount.toString(), BYTESIZES.AMOUNT, 'little');
        const votesBuffer = stringToBuffer(this.asset.votes.join(''));
        return Buffer.concat([
            transactionType,
            transactionTimestamp,
            transactionSenderPublicKey,
            transactionRecipientID,
            transactionAmount,
            votesBuffer,
        ]);
    }
    async prepare(store) {
        const publicKeyObjectArray = this.asset.votes.map(pkWithAction => {
            const publicKey = pkWithAction.slice(1);
            return {
                publicKey,
            };
        });
        const filterArray = [
            {
                address: this.senderId,
            },
            ...publicKeyObjectArray,
        ];
        await store.account.cache(filterArray);
    }
    verifyAgainstTransactions(transactions) {
        const sameTypeTransactions = transactions
            .filter(tx => tx.senderPublicKey === this.senderPublicKey && tx.type === this.type)
            .map(tx => new VoteTransaction(tx));
        const publicKeys = this.asset.votes.map(vote => vote.substring(1));
        return sameTypeTransactions.reduce((previous, tx) => {
            const conflictingVotes = tx.asset.votes
                .map(vote => vote.substring(1))
                .filter(publicKey => publicKeys.includes(publicKey));
            if (conflictingVotes.length > 0) {
                return [
                    ...previous,
                    new TransactionError(`Transaction includes conflicting votes: ${conflictingVotes.toString()}`, this.id, '.asset.votes'),
                ];
            }
            return previous;
        }, []);
    }
    validateAsset() {
        const asset = this.assetToJSON();
        const schemaErrors = validator.validate(exports.voteAssetFormatSchema, asset);
        const errors = convertToAssetError(this.id, schemaErrors);
        return errors;
    }
    applyAsset(store) {
        const errors = [];
        const sender = store.account.get(this.senderId);
        const balanceError = verifyAmountBalance(this.id, sender, this.asset.amount, this.fee);
        if (balanceError) {
            errors.push(balanceError);
        }
        const updatedSenderBalance = new bignum_1.default(sender.balance).sub(this.asset.amount);
        this.asset.votes.forEach(actionVotes => {
            const vote = actionVotes.substring(1);
            const voteAccount = store.account.find(account => account.publicKey === vote);
            if (!voteAccount ||
                (voteAccount &&
                    (voteAccount.username === undefined ||
                        voteAccount.username === '' ||
                        voteAccount.username === null))) {
                errors.push(new TransactionError(`${vote} is not a delegate.`, this.id, '.asset.votes'));
            }
        });
        const senderVotes = sender.votedDelegatesPublicKeys || [];
        this.asset.votes.forEach(vote => {
            const action = vote.charAt(0);
            const publicKey = vote.substring(1);
            if (action === PREFIX_UPVOTE && senderVotes.includes(publicKey)) {
                errors.push(new TransactionError(`${publicKey} is already voted.`, this.id, '.asset.votes'));
            }
            else if (action === PREFIX_UNVOTE && !senderVotes.includes(publicKey)) {
                errors.push(new TransactionError(`${publicKey} is not voted.`, this.id, '.asset.votes'));
            }
        });
        const upvotes = this.asset.votes
            .filter(vote => vote.charAt(0) === PREFIX_UPVOTE)
            .map(vote => vote.substring(1));
        const unvotes = this.asset.votes
            .filter(vote => vote.charAt(0) === PREFIX_UNVOTE)
            .map(vote => vote.substring(1));
        const originalVotes = sender.votedDelegatesPublicKeys || [];
        const votedDelegatesPublicKeys = [
            ...originalVotes,
            ...upvotes,
        ].filter(vote => !unvotes.includes(vote));
        if (votedDelegatesPublicKeys.length > MAX_VOTE_PER_ACCOUNT) {
            errors.push(new TransactionError(`Vote cannot exceed ${MAX_VOTE_PER_ACCOUNT} but has ${votedDelegatesPublicKeys.length}.`, this.id, '.asset.votes', votedDelegatesPublicKeys.length.toString(), MAX_VOTE_PER_ACCOUNT));
        }
        const updatedSender = {
            ...sender,
            balance: updatedSenderBalance.toString(),
            votedDelegatesPublicKeys,
        };
        store.account.set(updatedSender.address, updatedSender);
        return errors;
    }
    undoAsset(store) {
        const errors = [];
        const sender = store.account.get(this.senderId);
        const updatedSenderBalance = new bignum_1.default(sender.balance).add(this.asset.amount);
        if (updatedSenderBalance.gt(MAX_TRANSACTION_AMOUNT)) {
            errors.push(new TransactionError('Invalid amount', this.id, '.amount', this.asset.amount.toString()));
        }
        const upvotes = this.asset.votes
            .filter(vote => vote.charAt(0) === PREFIX_UPVOTE)
            .map(vote => vote.substring(1));
        const unvotes = this.asset.votes
            .filter(vote => vote.charAt(0) === PREFIX_UNVOTE)
            .map(vote => vote.substring(1));
        const originalVotes = sender.votedDelegatesPublicKeys || [];
        const votedDelegatesPublicKeys = [
            ...originalVotes,
            ...unvotes,
        ].filter(vote => !upvotes.includes(vote));
        if (votedDelegatesPublicKeys.length > MAX_VOTE_PER_ACCOUNT) {
            errors.push(new TransactionError(`Vote cannot exceed ${MAX_VOTE_PER_ACCOUNT} but has ${votedDelegatesPublicKeys.length}.`, this.id, '.asset.votes', votedDelegatesPublicKeys.length.toString(), MAX_VOTE_PER_ACCOUNT));
        }
        const updatedSender = {
            ...sender,
            balance: updatedSenderBalance.toString(),
            votedDelegatesPublicKeys,
        };
        store.account.set(updatedSender.address, updatedSender);
        return errors;
    }
}
exports.VoteTransaction = VoteTransaction;
VoteTransaction.TYPE = 3;
VoteTransaction.FEE = VOTE_FEE.toString();
//# sourceMappingURL=3_vote_transaction.js.map