"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bignum_1 = __importDefault(require("@liskhq/bignum"));
const lisk_sdk_1 = require("lisk-sdk");
const legacy_base_transaction_1 = require("./legacy_base_transaction");
const constants_1 = require("./constants");
const { convertToAssetError, TransactionError, utils: { convertBeddowsToLSK, verifyAmountBalance, }, constants, } = lisk_sdk_1.transactions;
const { isPositiveNumberString, validator, } = lisk_sdk_1.validator;
exports.inTransferAssetFormatSchema = {
    type: 'object',
    required: ['inTransfer', 'amount'],
    properties: {
        amount: {
            type: 'string',
            format: 'amount',
        },
        inTransfer: {
            type: 'object',
            required: ['dappId'],
            properties: {
                dappId: {
                    type: 'string',
                    format: 'id',
                },
            },
        },
    },
};
class InTransferTransaction extends legacy_base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        const rawAsset = tx.asset;
        this.asset = {
            amount: new bignum_1.default(isPositiveNumberString(rawAsset.amount) ? rawAsset.amount : '0'),
            inTransfer: rawAsset.inTransfer || {},
        };
    }
    getBasicBytes() {
        const transactionType = lisk_sdk_1.cryptography.intToBuffer(this.type, constants.BYTESIZES.TYPE);
        const transactionTimestamp = lisk_sdk_1.cryptography.intToBuffer(this.timestamp, constants.BYTESIZES.TIMESTAMP, 'little');
        const transactionSenderPublicKey = lisk_sdk_1.cryptography.hexToBuffer(this.senderPublicKey);
        const transactionRecipientID = Buffer.alloc(constants.BYTESIZES.RECIPIENT_ID);
        const transactionAmount = lisk_sdk_1.cryptography.bigNumberToBuffer(this.asset.amount.toString(), constants.BYTESIZES.AMOUNT, 'little');
        return Buffer.concat([
            transactionType,
            transactionTimestamp,
            transactionSenderPublicKey,
            transactionRecipientID,
            transactionAmount,
            Buffer.from(this.asset.inTransfer.dappId, 'utf8'),
        ]);
    }
    async prepare(store) {
        await store.account.cache([{ address: this.senderId }]);
        const transactions = await store.transaction.cache([
            {
                id: this.asset.inTransfer.dappId,
            },
        ]);
        const dappTransaction = transactions && transactions.length > 0
            ? transactions.find(tx => tx.type === constants_1.TRANSACTION_DAPP_TYPE &&
                tx.id === this.asset.inTransfer.dappId)
            : undefined;
        if (dappTransaction) {
            await store.account.cache([
                { address: lisk_sdk_1.cryptography.getAddressFromPublicKey(dappTransaction.senderPublicKey) },
            ]);
        }
    }
    assetToJSON() {
        return {
            ...this.asset,
            amount: this.asset.amount.toString(),
        };
    }
    verifyAgainstTransactions(_) {
        return [];
    }
    validateAsset() {
        const asset = this.assetToJSON();
        const schemaErrors = validator.validate(exports.inTransferAssetFormatSchema, asset);
        const errors = convertToAssetError(this.id, schemaErrors);
        if (this.asset.amount.lte(0)) {
            errors.push(new TransactionError('Amount must be greater than 0', this.id, '.amount', this.asset.amount.toString(), '0'));
        }
        return errors;
    }
    applyAsset(store) {
        const errors = [];
        const idExists = store.transaction.find((transaction) => transaction.type === constants_1.TRANSACTION_DAPP_TYPE &&
            transaction.id === this.asset.inTransfer.dappId);
        if (!idExists) {
            errors.push(new TransactionError(`Application not found: ${this.asset.inTransfer.dappId}`, this.id, this.asset.inTransfer.dappId));
        }
        const sender = store.account.get(this.senderId);
        const balanceError = verifyAmountBalance(this.id, sender, this.asset.amount, this.fee);
        if (balanceError) {
            errors.push(balanceError);
        }
        const updatedBalance = new bignum_1.default(sender.balance).sub(this.asset.amount);
        const updatedSender = { ...sender, balance: updatedBalance.toString() };
        store.account.set(updatedSender.address, updatedSender);
        const dappTransaction = store.transaction.get(this.asset.inTransfer.dappId);
        const recipient = store.account.get(lisk_sdk_1.cryptography.getAddressFromPublicKey(dappTransaction.senderPublicKey));
        const updatedRecipientBalance = new bignum_1.default(recipient.balance).add(this.asset.amount);
        const updatedRecipient = {
            ...recipient,
            balance: updatedRecipientBalance.toString(),
        };
        store.account.set(updatedRecipient.address, updatedRecipient);
        return errors;
    }
    undoAsset(store) {
        const errors = [];
        const sender = store.account.get(this.senderId);
        const updatedBalance = new bignum_1.default(sender.balance).add(this.asset.amount);
        const updatedSender = { ...sender, balance: updatedBalance.toString() };
        store.account.set(updatedSender.address, updatedSender);
        const dappTransaction = store.transaction.get(this.asset.inTransfer.dappId);
        const recipient = store.account.get(lisk_sdk_1.cryptography.getAddressFromPublicKey(dappTransaction.senderPublicKey));
        const updatedRecipientBalance = new bignum_1.default(recipient.balance).sub(this.asset.amount);
        if (updatedRecipientBalance.lt(0)) {
            errors.push(new TransactionError(`Account does not have enough LSK: ${recipient.address}, balance: ${convertBeddowsToLSK(recipient.balance)}.`, this.id));
        }
        const updatedRecipient = {
            ...recipient,
            balance: updatedRecipientBalance.toString(),
        };
        store.account.set(updatedRecipient.address, updatedRecipient);
        return errors;
    }
    assetFromSync(raw) {
        if (!raw.in_dappId) {
            return undefined;
        }
        const inTransfer = {
            dappId: raw.in_dappId,
        };
        return { inTransfer };
    }
}
exports.InTransferTransaction = InTransferTransaction;
InTransferTransaction.TYPE = 6;
InTransferTransaction.FEE = constants_1.IN_TRANSFER_FEE.toString();
//# sourceMappingURL=6_in_transfer_transaction.js.map