"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bignum_1 = __importDefault(require("@liskhq/bignum"));
const lisk_sdk_1 = require("lisk-sdk");
const legacy_base_transaction_1 = require("./legacy_base_transaction");
const constants_1 = require("./constants");
const { convertToAssetError, TransactionError, utils: { verifyAmountBalance, }, constants, } = lisk_sdk_1.transactions;
const { isPositiveNumberString, validator, } = lisk_sdk_1.validator;
const TRANSACTION_DAPP_REGISTERATION_TYPE = 5;
exports.outTransferAssetFormatSchema = {
    type: 'object',
    required: ['outTransfer', 'recipientId', 'amount'],
    properties: {
        recipientId: {
            type: 'string',
            format: 'address',
        },
        amount: {
            type: 'string',
            format: 'amount',
        },
        outTransfer: {
            type: 'object',
            required: ['dappId', 'transactionId'],
            properties: {
                dappId: {
                    type: 'string',
                    format: 'id',
                },
                transactionId: {
                    type: 'string',
                    format: 'id',
                },
            },
        },
    },
};
class OutTransferTransaction extends legacy_base_transaction_1.BaseTransaction {
    constructor(rawTransaction) {
        super(rawTransaction);
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        const rawAsset = tx.asset;
        this.asset = {
            amount: new bignum_1.default(isPositiveNumberString(rawAsset.amount) ? rawAsset.amount : '0'),
            recipientId: rawAsset.recipientId || '',
            outTransfer: rawAsset.outTransfer || {},
        };
        this.containsUniqueData = true;
    }
    async prepare(store) {
        await store.account.cache([
            {
                address: this.senderId,
            },
            { address: this.asset.recipientId },
        ]);
        await store.transaction.cache([
            {
                id: this.asset.outTransfer.dappId,
            },
            { id: this.asset.outTransfer.transactionId },
        ]);
    }
    getBasicBytes() {
        const transactionType = lisk_sdk_1.cryptography.intToBuffer(this.type, constants.BYTESIZES.TYPE);
        const transactionTimestamp = lisk_sdk_1.cryptography.intToBuffer(this.timestamp, constants.BYTESIZES.TIMESTAMP, 'little');
        const transactionSenderPublicKey = lisk_sdk_1.cryptography.hexToBuffer(this.senderPublicKey);
        const transactionRecipientID = lisk_sdk_1.cryptography.intToBuffer(this.asset.recipientId.slice(0, -1), constants.BYTESIZES.RECIPIENT_ID).slice(0, constants.BYTESIZES.RECIPIENT_ID);
        const transactionAmount = lisk_sdk_1.cryptography.bigNumberToBuffer(this.asset.amount.toString(), constants.BYTESIZES.AMOUNT, 'little');
        const { dappId, transactionId } = this.asset.outTransfer;
        const outAppIdBuffer = Buffer.from(dappId, 'utf8');
        const outTransactionIdBuffer = Buffer.from(transactionId, 'utf8');
        return Buffer.concat([
            transactionType,
            transactionTimestamp,
            transactionSenderPublicKey,
            transactionRecipientID,
            transactionAmount,
            Buffer.concat([outAppIdBuffer, outTransactionIdBuffer]),
        ]);
    }
    assetToJSON() {
        return {
            ...this.asset,
            amount: this.asset.amount.toString(),
        };
    }
    verifyAgainstTransactions(transactions) {
        const sameTypeTransactions = transactions.filter(tx => tx.type === OutTransferTransaction.TYPE &&
            'outTransfer' in tx.asset &&
            this.asset.outTransfer.transactionId ===
                tx.asset.outTransfer.transactionId);
        return sameTypeTransactions.length > 0
            ? [
                new TransactionError('Out Transfer cannot refer to the same transactionId', this.id, '.asset.outTransfer.transactionId'),
            ]
            : [];
    }
    validateAsset() {
        const asset = this.assetToJSON();
        const schemaErrors = validator.validate(exports.outTransferAssetFormatSchema, asset);
        const errors = convertToAssetError(this.id, schemaErrors);
        if (this.asset.amount.lte(0)) {
            errors.push(new TransactionError('Amount must be greater than zero for outTransfer transaction', this.id, '.amount', this.asset.amount.toString()));
        }
        if (this.asset.recipientId === '') {
            errors.push(new TransactionError('RecipientId must be set for outTransfer transaction', this.id, '.recipientId', this.asset.recipientId));
        }
        return errors;
    }
    applyAsset(store) {
        const errors = [];
        const dappRegistrationTransaction = store.transaction.get(this.asset.outTransfer.dappId);
        if (!dappRegistrationTransaction ||
            dappRegistrationTransaction.type !== TRANSACTION_DAPP_REGISTERATION_TYPE) {
            errors.push(new TransactionError(`Application not found: ${this.asset.outTransfer.dappId}`, this.id, '.asset.outTransfer.dappId'));
        }
        const sender = store.account.get(this.senderId);
        const balanceError = verifyAmountBalance(this.id, sender, this.asset.amount, this.fee);
        if (balanceError) {
            errors.push(balanceError);
        }
        const updatedBalance = new bignum_1.default(sender.balance).sub(this.asset.amount);
        const updatedSender = { ...sender, balance: updatedBalance.toString() };
        store.account.set(updatedSender.address, updatedSender);
        const recipient = store.account.getOrDefault(this.asset.recipientId);
        const updatedRecipientBalance = new bignum_1.default(recipient.balance).add(this.asset.amount);
        if (updatedRecipientBalance.gt(constants_1.MAX_TRANSACTION_AMOUNT)) {
            errors.push(new TransactionError('Invalid amount', this.id, '.amount'));
        }
        const updatedRecipient = {
            ...recipient,
            balance: updatedRecipientBalance.toString(),
        };
        store.account.set(updatedRecipient.address, updatedRecipient);
        return errors;
    }
    undoAsset(store) {
        const errors = [];
        const sender = store.account.get(this.senderId);
        const updatedBalance = new bignum_1.default(sender.balance).add(this.asset.amount);
        if (updatedBalance.gt(constants_1.MAX_TRANSACTION_AMOUNT)) {
            errors.push(new TransactionError('Invalid amount', this.id, '.amount', this.asset.amount.toString()));
        }
        const updatedSender = { ...sender, balance: updatedBalance.toString() };
        store.account.set(updatedSender.address, updatedSender);
        const recipient = store.account.getOrDefault(this.asset.recipientId);
        const updatedRecipientBalance = new bignum_1.default(recipient.balance).sub(this.asset.amount);
        if (updatedRecipientBalance.lt(0)) {
            errors.push(new TransactionError(`Account does not have enough LSK: ${recipient.address}, balance: ${recipient.balance}`, this.id, updatedRecipientBalance.toString()));
        }
        const updatedRecipient = {
            ...recipient,
            balance: updatedRecipientBalance.toString(),
        };
        store.account.set(updatedRecipient.address, updatedRecipient);
        return errors;
    }
    assetFromSync(raw) {
        if (!raw.ot_dappId) {
            return undefined;
        }
        const outTransfer = {
            dappId: raw.ot_dappId,
            transactionId: raw.ot_outTransactionId,
        };
        return { outTransfer };
    }
}
exports.OutTransferTransaction = OutTransferTransaction;
OutTransferTransaction.TYPE = 7;
OutTransferTransaction.FEE = constants_1.OUT_TRANSFER_FEE.toString();
//# sourceMappingURL=7_out_transfer_transaction.js.map