"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const bignum_1 = __importDefault(require("@liskhq/bignum"));
const lisk_sdk_1 = require("lisk-sdk");
const schemas = __importStar(require("./schema"));
const { TransactionError, TransactionPendingError, createResponse, convertToTransactionError, Status, utils: { getId, validateSenderIdAndPublicKey, validateSignature, verifyBalance, verifyMultiSignatures, verifySecondSignature, verifySenderPublicKey, }, constants: { BYTESIZES, MAX_TRANSACTION_AMOUNT, UNCONFIRMED_MULTISIG_TRANSACTION_TIMEOUT, UNCONFIRMED_TRANSACTION_TIMEOUT, }, } = lisk_sdk_1.transactions;
const { validator, } = lisk_sdk_1.validator;
const { getAddressAndPublicKeyFromPassphrase, getAddressFromPublicKey, hash, hexToBuffer, signData, } = lisk_sdk_1.cryptography;
var MultisignatureStatus;
(function (MultisignatureStatus) {
    MultisignatureStatus[MultisignatureStatus["UNKNOWN"] = 0] = "UNKNOWN";
    MultisignatureStatus[MultisignatureStatus["NONMULTISIGNATURE"] = 1] = "NONMULTISIGNATURE";
    MultisignatureStatus[MultisignatureStatus["PENDING"] = 2] = "PENDING";
    MultisignatureStatus[MultisignatureStatus["READY"] = 3] = "READY";
    MultisignatureStatus[MultisignatureStatus["FAIL"] = 4] = "FAIL";
})(MultisignatureStatus = exports.MultisignatureStatus || (exports.MultisignatureStatus = {}));
exports.ENTITY_ACCOUNT = 'account';
exports.ENTITY_TRANSACTION = 'transaction';
class BaseTransaction {
    constructor(rawTransaction) {
        this._multisignatureStatus = MultisignatureStatus.UNKNOWN;
        const tx = (typeof rawTransaction === 'object' && rawTransaction !== null
            ? rawTransaction
            : {});
        this.fee = new bignum_1.default(this.constructor.FEE);
        this.type =
            typeof tx.type === 'number'
                ? tx.type
                : this.constructor.TYPE;
        this._id = tx.id;
        this._senderPublicKey = tx.senderPublicKey || '';
        this._signature = tx.signature;
        this.signatures = tx.signatures || [];
        this._signSignature = tx.signSignature;
        this.timestamp = typeof tx.timestamp === 'number' ? tx.timestamp : 0;
        this.confirmations = tx.confirmations;
        this.blockId = tx.blockId;
        this.height = tx.height;
        this.receivedAt = tx.receivedAt ? new Date(tx.receivedAt) : undefined;
        this.asset = tx.asset || {};
    }
    get id() {
        return this._id || '';
    }
    get senderId() {
        return getAddressFromPublicKey(this.senderPublicKey);
    }
    get senderPublicKey() {
        if (!this._senderPublicKey) {
            throw new Error('senderPublicKey is required to be set before use');
        }
        return this._senderPublicKey;
    }
    get signature() {
        if (!this._signature) {
            throw new Error('signature is required to be set before use');
        }
        return this._signature;
    }
    get signSignature() {
        return this._signSignature;
    }
    toJSON() {
        const transaction = {
            id: this.id,
            blockId: this.blockId,
            height: this.height,
            confirmations: this.confirmations,
            type: this.type,
            timestamp: this.timestamp,
            senderPublicKey: this._senderPublicKey || '',
            senderId: this._senderPublicKey ? this.senderId : '',
            fee: this.fee.toString(),
            signature: this._signature,
            signSignature: this.signSignature ? this.signSignature : undefined,
            signatures: this.signatures,
            asset: this.assetToJSON(),
            receivedAt: this.receivedAt ? this.receivedAt.toISOString() : undefined,
        };
        return transaction;
    }
    stringify() {
        return JSON.stringify(this.toJSON());
    }
    isReady() {
        return (this._multisignatureStatus === MultisignatureStatus.READY ||
            this._multisignatureStatus === MultisignatureStatus.NONMULTISIGNATURE);
    }
    getBytes() {
        const transactionBytes = Buffer.concat([
            this.getBasicBytes(),
            this._signature ? hexToBuffer(this._signature) : Buffer.alloc(0),
            this._signSignature ? hexToBuffer(this._signSignature) : Buffer.alloc(0),
        ]);
        return transactionBytes;
    }
    validate() {
        const errors = [...this._validateSchema(), ...this.validateAsset()];
        if (errors.length > 0) {
            return createResponse(this.id, errors);
        }
        const transactionBytes = this.getBasicBytes();
        this._id = getId(this.getBytes());
        const { valid: signatureValid, error: verificationError, } = validateSignature(this.senderPublicKey, this.signature, transactionBytes, this.id);
        if (!signatureValid && verificationError) {
            errors.push(verificationError);
        }
        if (this.type !== this.constructor.TYPE) {
            errors.push(new TransactionError(`Invalid type`, this.id, '.type', this.type, this.constructor.TYPE));
        }
        return createResponse(this.id, errors);
    }
    verifyAgainstTransactions(_) {
        return [];
    }
    verifyAgainstOtherTransactions(transactions) {
        const errors = this.verifyAgainstTransactions(transactions);
        return createResponse(this.id, errors);
    }
    apply(store) {
        const sender = store.account.getOrDefault(this.senderId);
        const errors = this._verify(sender);
        const { errors: multiSigError } = this.processMultisignatures(store);
        if (multiSigError) {
            errors.push(...multiSigError);
        }
        const updatedBalance = new bignum_1.default(sender.balance).sub(this.fee);
        const updatedSender = {
            ...sender,
            balance: updatedBalance.toString(),
            publicKey: sender.publicKey || this.senderPublicKey,
        };
        store.account.set(updatedSender.address, updatedSender);
        const assetErrors = this.applyAsset(store);
        errors.push(...assetErrors);
        if (this._multisignatureStatus === MultisignatureStatus.PENDING &&
            errors.length === 1 &&
            errors[0] instanceof TransactionPendingError) {
            return {
                id: this.id,
                status: Status.PENDING,
                errors,
            };
        }
        return createResponse(this.id, errors);
    }
    undo(store) {
        const sender = store.account.getOrDefault(this.senderId);
        const updatedBalance = new bignum_1.default(sender.balance).add(this.fee);
        const updatedAccount = {
            ...sender,
            balance: updatedBalance.toString(),
            publicKey: sender.publicKey || this.senderPublicKey,
        };
        const errors = updatedBalance.lte(MAX_TRANSACTION_AMOUNT)
            ? []
            : [
                new TransactionError('Invalid balance amount', this.id, '.balance', sender.balance, updatedBalance.toString()),
            ];
        store.account.set(updatedAccount.address, updatedAccount);
        const assetErrors = this.undoAsset(store);
        errors.push(...assetErrors);
        return createResponse(this.id, errors);
    }
    async prepare(store) {
        await store.account.cache([
            {
                address: this.senderId,
            },
        ]);
    }
    addMultisignature(store, signatureObject) {
        const account = store.account.get(this.senderId);
        if (account.membersPublicKeys &&
            !account.membersPublicKeys.includes(signatureObject.publicKey)) {
            return createResponse(this.id, [
                new TransactionError(`Public Key '${signatureObject.publicKey}' is not a member for account '${account.address}'.`, this.id),
            ]);
        }
        if (this.signatures.includes(signatureObject.signature)) {
            return createResponse(this.id, [
                new TransactionError(`Signature '${signatureObject.signature}' already present in transaction.`, this.id),
            ]);
        }
        const { valid } = validateSignature(signatureObject.publicKey, signatureObject.signature, this.getBasicBytes(), this.id);
        if (valid) {
            this.signatures.push(signatureObject.signature);
            return this.processMultisignatures(store);
        }
        const errors = valid
            ? []
            : [
                new TransactionError(`Failed to add signature '${signatureObject.signature}'.`, this.id, '.signatures'),
            ];
        return createResponse(this.id, errors);
    }
    addVerifiedMultisignature(signature) {
        if (!this.signatures.includes(signature)) {
            this.signatures.push(signature);
            return createResponse(this.id, []);
        }
        return createResponse(this.id, [
            new TransactionError('Failed to add signature.', this.id, '.signatures'),
        ]);
    }
    processMultisignatures(store) {
        const sender = store.account.get(this.senderId);
        const transactionBytes = this.getBasicBytes();
        const { status, errors } = verifyMultiSignatures(this.id, sender, this.signatures, transactionBytes);
        this._multisignatureStatus = status;
        if (this._multisignatureStatus === MultisignatureStatus.PENDING) {
            return {
                id: this.id,
                status: Status.PENDING,
                errors,
            };
        }
        return createResponse(this.id, errors);
    }
    isExpired(date = new Date()) {
        if (!this.receivedAt) {
            this.receivedAt = new Date();
        }
        const timeNow = Math.floor(date.getTime() / 1000);
        const timeOut = this._multisignatureStatus === MultisignatureStatus.PENDING ||
            this._multisignatureStatus === MultisignatureStatus.READY
            ? UNCONFIRMED_MULTISIG_TRANSACTION_TIMEOUT
            : UNCONFIRMED_TRANSACTION_TIMEOUT;
        const timeElapsed = timeNow - Math.floor(this.receivedAt.getTime() / 1000);
        return timeElapsed > timeOut;
    }
    sign(passphrase, secondPassphrase) {
        const { publicKey } = getAddressAndPublicKeyFromPassphrase(passphrase);
        if (this._senderPublicKey !== '' && this._senderPublicKey !== publicKey) {
            throw new Error('Transaction senderPublicKey does not match public key from passphrase');
        }
        this._senderPublicKey = publicKey;
        this._signature = undefined;
        this._signSignature = undefined;
        this._signature = signData(hash(this.getBytes()), passphrase);
        if (secondPassphrase) {
            this._signSignature = signData(hash(this.getBytes()), secondPassphrase);
        }
        this._id = getId(this.getBytes());
    }
    getBasicBytes() {
        const transactionType = Buffer.alloc(BYTESIZES.TYPE, this.type);
        const transactionTimestamp = Buffer.alloc(BYTESIZES.TIMESTAMP);
        transactionTimestamp.writeIntLE(this.timestamp, 0, BYTESIZES.TIMESTAMP);
        const transactionSenderPublicKey = hexToBuffer(this.senderPublicKey);
        const transactionRecipientID = Buffer.alloc(BYTESIZES.RECIPIENT_ID);
        const transactionAmount = new bignum_1.default(0).toBuffer({
            endian: 'little',
            size: BYTESIZES.AMOUNT,
        });
        return Buffer.concat([
            transactionType,
            transactionTimestamp,
            transactionSenderPublicKey,
            transactionRecipientID,
            transactionAmount,
            this.assetToBytes(),
        ]);
    }
    assetToJSON() {
        return this.asset;
    }
    assetToBytes() {
        return Buffer.from(JSON.stringify(this.asset), 'utf-8');
    }
    _verify(sender) {
        const secondSignatureTxBytes = Buffer.concat([
            this.getBasicBytes(),
            hexToBuffer(this.signature),
        ]);
        return [
            verifySenderPublicKey(this.id, sender, this.senderPublicKey),
            verifyBalance(this.id, sender, this.fee),
            verifySecondSignature(this.id, sender, this.signSignature, secondSignatureTxBytes),
        ].filter(Boolean);
    }
    _validateSchema() {
        const transaction = this.toJSON();
        const schemaErrors = validator.validate(schemas.baseTransaction, transaction);
        const errors = convertToTransactionError(this.id, schemaErrors);
        if (!errors.find((err) => err.dataPath === '.senderPublicKey')) {
            const senderIdError = validateSenderIdAndPublicKey(this.id, this.senderId, this.senderPublicKey);
            if (senderIdError) {
                errors.push(senderIdError);
            }
        }
        return errors;
    }
}
exports.BaseTransaction = BaseTransaction;
BaseTransaction.FEE = '0';
//# sourceMappingURL=legacy_base_transaction.js.map