"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const command_1 = require("@oclif/command");
const base_1 = tslib_1.__importDefault(require("../../base"));
const error_1 = require("../../utils/error");
const flags_1 = require("../../utils/flags");
const input_1 = require("../../utils/input");
const utils_1 = require("../../utils/input/utils");
const network_identifier_1 = require("../../utils/network_identifier");
const object_1 = require("../../utils/object");
const transactions_1 = require("../../utils/transactions");
const getTransactionInput = async () => {
    try {
        const { data } = await utils_1.getStdIn({ dataIsRequired: true });
        if (!data) {
            throw new error_1.ValidationError('No transaction was provided.');
        }
        return data;
    }
    catch (e) {
        throw new error_1.ValidationError('No transaction was provided.');
    }
};
class SignCommand extends base_1.default {
    async run() {
        const { args, flags: { networkIdentifier: networkIdentifierSource, passphrase: passphraseSource, 'second-passphrase': secondPassphraseSource, }, } = this.parse(SignCommand);
        const { transaction } = args;
        const transactionInput = transaction || (await getTransactionInput());
        const transactionObject = transactions_1.parseTransactionString(transactionInput);
        const { passphrase, secondPassphrase } = await input_1.getInputsFromSources({
            passphrase: {
                source: passphraseSource,
                repeatPrompt: true,
            },
            secondPassphrase: !secondPassphraseSource
                ? undefined
                : {
                    source: secondPassphraseSource,
                    repeatPrompt: true,
                },
        });
        if (!passphrase) {
            throw new Error('Passphrase is required to sign the transaction');
        }
        const networkIdentifier = network_identifier_1.getNetworkIdentifierWithInput(networkIdentifierSource, this.userConfig.api.network);
        const txInstance = transactions_1.instantiateTransaction({
            ...transactionObject,
            networkIdentifier,
        });
        txInstance.sign(passphrase, secondPassphrase);
        const { errors } = txInstance.validate();
        if (errors.length !== 0) {
            throw errors;
        }
        this.print(object_1.removeUndefinedValues(txInstance.toJSON()));
    }
}
exports.default = SignCommand;
SignCommand.args = [
    {
        name: 'transaction',
        description: 'Transaction to sign in JSON format.',
    },
];
SignCommand.description = `
	Sign a transaction using your secret passphrase.
	`;
SignCommand.examples = [
    'transaction:sign \'{"amount":"100","recipientId":"13356260975429434553L","senderPublicKey":null,"timestamp":52871598,"type":0,"fee":"10000000","recipientPublicKey":null,"asset":{}}\'',
];
SignCommand.flags = {
    ...base_1.default.flags,
    networkIdentifier: command_1.flags.string(flags_1.flags.networkIdentifier),
    passphrase: command_1.flags.string(flags_1.flags.passphrase),
    'second-passphrase': command_1.flags.string(flags_1.flags.secondPassphrase),
};
//# sourceMappingURL=sign.js.map